/***

The MIT License

Copyright (c) 2018 Teppei Tamra (TAM)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

***/

#include "saoridon.h"
#include "saoridef.h"
#include <QtNetwork/QNetworkAccessManager>
#include <QtNetwork/QNetworkReply>
#include <QJsonDocument>
#include <QJsonObject>
#include <QEventLoop>
#include <QUrlQuery>
#include <QDebug>

Saoridon::Saoridon(QUrl instanceUrl,QObject *parent) : QObject(parent)
{
    setInstance(instanceUrl);
    getInstanceInfomation();
}

const QUrl Saoridon::instance()
{
    return m_instance;
}

void Saoridon::setInstance(const QUrl instanceUrl)
{
    m_instance = instanceUrl;
    return;
}

const QString Saoridon::clientId()
{
    return m_clientId;
}

void Saoridon::setClientId(const QString id)
{
    m_clientId = id;
    return;
}

const QString Saoridon::clientSecret()
{
    return m_clientSecret;
}

void Saoridon::setClientSecret(const QString secret)
{
    m_clientSecret = secret;
}

bool Saoridon::clientRedistration()
{
    auto manager = new QNetworkAccessManager();
    QEventLoop event;
    connect(manager,SIGNAL(finished(QNetworkReply*)),&event,SLOT(quit()));
    QUrlQuery params;
    params.addQueryItem("client_name",SAORI_CLIENT_NAME);
    params.addQueryItem("redirect_uris",SAORI_CLIENT_REDIRECT_URI);
    params.addQueryItem("scopes",SAORI_CLIENT_SCOPES);
    params.addQueryItem("website",SAORI_CLIENT_WEBSITE);
    QNetworkRequest request;
    request.setHeader(QNetworkRequest::ContentTypeHeader,QVariant("application/x-www-form-urlencoded"));
    request.setUrl(QUrl(instance().url() + SAORI_MASTODON_APIPATH_APPS));
    auto *reply = manager->post(request,params.toString().toLatin1());
    event.exec();
    if (reply->error() != QNetworkReply::NoError) {
        manager->deleteLater();
        return false;
    }
    QJsonObject json = QJsonDocument::fromJson(reply->readAll()).object();
    QString id = json["client_id"].toString();
    QString secret = json["client_secret"].toString();
    if (id.isEmpty() || secret.isEmpty()) return false;
    setClientId(id);
    setClientSecret(secret);
    manager->deleteLater();
    return true;
}

const QUrl Saoridon::getAuthorizedUrl()
{
    //if (clientId().isEmpty() || clientSecret().isEmpty()) return QUrl();
    QUrlQuery params;
    params.addQueryItem("client_id",clientId());
    params.addQueryItem("response_type","code");
    params.addQueryItem("redirect_uri","urn:ietf:wg:oauth:2.0:oob");
    params.addQueryItem("scope",SAORI_CLIENT_SCOPES);
    QUrl url = instance().url() + SAORI_MASTODON_APIPATH_AUTHORIZE;
    url.setQuery(params);
    return url;



}

const QString Saoridon::getAccessToken(const QString code)
{
    auto manager = new QNetworkAccessManager();
    QEventLoop event;
    connect(manager,&QNetworkAccessManager::finished,&event,&QEventLoop::quit);
    QUrlQuery params;
    params.addQueryItem("grant_type","authorization_code");
    params.addQueryItem("redirect_uri","urn:ietf:wg:oauth:2.0:oob");
    params.addQueryItem("client_id",clientId());
    params.addQueryItem("client_secret",clientSecret());
    params.addQueryItem("code",code);
    QNetworkRequest request;
    request.setHeader(QNetworkRequest::ContentTypeHeader,QVariant("application/x-www-form-urlencoded"));
    request.setUrl(QUrl(instance().url() + SAORI_MASTODON_APIPATH_TOKEN));
    auto *reply = manager->post(request,params.toString().toLatin1());
    event.exec();
    if (reply->error() != QNetworkReply::NoError) {
        manager->deleteLater();
        return QString();
    }
    QJsonObject json = QJsonDocument::fromJson(reply->readAll()).object();
    QString token = json["access_token"].toString();
    manager->deleteLater();
    return token;
}

void Saoridon::getInstanceInfomation()
{
    auto manager = new QNetworkAccessManager();
    QNetworkRequest request;
    request.setUrl(QUrl(instance().url() + SAORI_MASTODON_APIPATH_INSTANCE));
    auto *reply = manager->get(request);
    connect(manager,&QNetworkAccessManager::finished,[=](){
        if (reply->NoError == QNetworkReply::NoError) {
            QJsonObject json = QJsonDocument::fromJson(reply->readAll()).object();
            QMap<QString,QString> info;
            for (auto it = json.begin();it != json.end();it ++) {
                info[it.key()] = it.value().toString();
            }
            m_instanceInfo = info;
            emit instanceInfomationChanged();
        }
        manager->deleteLater();
    });

    m_timelineMap.clear();
    m_timelineMap["home"] = QUrl(instance().url() + SAORI_MASTODON_APIPATH_TIMELINE + "/home");
    m_timelineMap["local"] = QUrl(instance().url() + SAORI_MASTODON_APIPATH_TIMELINE + "/local");
    m_timelineMap["public"] = QUrl(instance().url() + SAORI_MASTODON_APIPATH_TIMELINE + "/public");
    m_timelineMap["notifications"] = QUrl(instance().url() + SAORI_MASTODON_APIPATH_TIMELINE + "/notifications");
    return;
}

const QString Saoridon::instanceInfo(const QString key)
{
    return m_instanceInfo[key];
}

const QUrl Saoridon::timelineUrl(const QString timeline)
{
    return m_timelineMap[timeline];
}

